// RTTY-Decoder from DD0CW with modifications by DJ7OO
// select between 45.45Baud,50Baud,75Baud,110Baud,300Baud by switch on #11
// select between normal and invers operation by switch on #12
// all settings will bestored in EEPROM
// data from decoder LM567 via port input #A1

#include <FlexiTimer2.h>   // Bibliothek für Timer Interrupts
#include <LiquidCrystal.h> // Bibliothek für LCD-Display
#include <EEPROM.h>        // Bibliothek für EEPROM Behandlung

LiquidCrystal lcd(8, 9, 4, 5, 6, 7); // LCD-Display 2 x 16 an entspr. Pins

byte     ti;  // Zähler Timer-millisekunden
byte     ti_speed[7]; // Zeitpunkte in ms zum Abtasten des seriellen Eing.-Signals
const byte     ti_45[7]  = {10, 22, 44, 66, 88, 110, 135}; // bei 45 Baud
const byte     ti_50[7]  = { 9, 20, 40, 60, 80, 100, 123}; //     50 Baud
const byte     ti_75[7]  = { 7, 13, 27, 40, 53,  67, 82};  //     75 Baud
const byte     ti_100[7] = { 5, 10, 20, 30, 40,  50, 62};  //    100 Baud
const byte     ti_110[7] = { 4,  9, 18, 27, 36,  46, 57};  //    110 Baud
const byte     ti_300[7] = { 1,  3,  7, 10, 13,  17, 21};  //    300 Baud

byte           inverse = 1; // 0 = Normal , 1 = Invers

boolean        fig = false; // Zahlen- oder Buchstaben-Ebene
byte           baud_rate;
boolean        sw_11 = false; // Taster an Digital-Pin 11 für Baud-Rate-Auswahl gedrückt ?
boolean        sw_12 = false; // Taster an Digital-Pin 12 für Normal/Invers & Baudot/ASCII gedrückt ?

uint8_t        baud_buff[10] = {0, 0, 0, 0, 0}; // 5 Zeichen Puffer für empf. Baudot-Zeichen
uint8_t        baud = 0;   // 1 Byte zum Zusammensetzen eines empfangenen Zeichens

///////////////////////////////////////////////////////////////
void timer_interrupt(void)
{
  static byte signal;

  static byte rSq;       // Abschnitt eines RTTY-Zeichens: 0=keine Daten, 1=Startbit, 2=Datenbits
  static byte sw11_ti;    // Timer für Taster an Pin 11
  static byte sw12_ti;    // Timer für Taster an Pin 12
  const byte min_press_ti = 100; // Taster muß mind. x millisek. gedrückt werden
  int i;

  if ( sw_11 == true || sw_12 == true ) {
    return;
  }
  // wenn einer der Taster noch als gedrückt gilt, dann hier nicht weiter machen.
  // Variable muß erst ausserhalb wieder zurück gesetzt werden.

  // Taster an Pin 11
  if ( digitalRead(11) == 1 && sw11_ti < min_press_ti ) // nur sehr kurz gedrückt und los gelassen bzw. geprellt
  {
    sw11_ti = 0;  // dann weiter machen mit Dekodieren wie normal, Taster nicht beachten
  }
  else if ( digitalRead(11) == 0 && sw11_ti < min_press_ti ) // noch immer gedrückt, aber noch nicht lange genug
  {
    sw11_ti++ ; return;
  }
  else if ( digitalRead(11) == 0 && sw11_ti >= min_press_ti ) // noch immer gedrückt, und auch schon lange genug
  {
    sw_11 = true;
    sw11_ti = ti = baud = 0; return;
  }

  // Taster an Pin 12
  if ( digitalRead(12) == 1 && sw12_ti < min_press_ti )
  {
    sw12_ti = 0;
  }
  else if ( digitalRead(12) == 0 && sw12_ti < min_press_ti )
  {
    sw12_ti++ ;
    return;
  }
  else if ( digitalRead(12) == 0 && sw12_ti >= min_press_ti )
  {
    sw_12 = true;
    sw12_ti = ti = baud = 0;
    return;
  }

  ti++;
  signal = digitalRead(15);    // = #A1
  if ( inverse == 1 ) {
    if ( signal == 0 ) {
      signal = 1;
    }
    else {
      signal = 0;
    }
  }

  if (rSq == 0 && signal == 0) {
    rSq = 1;
    ti = baud = 0;
  }
  else if (rSq == 1 && ti == ti_speed[0])
  {
    if (signal == 0)
    {
      rSq = 2;
      ti = baud = 0;
    }
    else
    {
      rSq = baud = 0;
    }
  }
  else if (rSq == 2 && ti == ti_speed[1])
  {
    bitWrite(baud, 0, signal);
  }
  else if (rSq == 2 && ti == ti_speed[2])
  {
    bitWrite(baud, 1, signal);
  }
  else if (rSq == 2 && ti == ti_speed[3])
  {
    bitWrite(baud, 2, signal);
  }
  else if (rSq == 2 && ti == ti_speed[4])
  {
    bitWrite(baud, 3, signal);
  }
  else if (rSq == 2 && ti == ti_speed[5])
  {
    bitWrite(baud, 4, signal);
    for ( i = 0; i <= 4 ; i++ ) {
      if ( baud_buff[i] == 0 ) {
        baud_buff[i] = baud;
        baud = 0; break; // auf den ersten freien Platz von links im Puffer eintragen
      }
    }
    // falls der Puffer voll ist, dann geht das vorherige Zeichen verloren
  }
  else if (rSq == 2 && ti >= ti_speed[6])
  {
    rSq = baud = 0;
  }
}

//////////////////////////////////////////////////////////////
void init_baud_timer()
{
  baud_rate = EEPROM.read(0);
  if      ( baud_rate == 45 )  {
    for ( int i = 0; i < 7; i++ ) {
      ti_speed[i] = ti_45[i] ;
    }
  }
  else if ( baud_rate == 50 )  {
    for ( int i = 0; i < 7; i++ ) {
      ti_speed[i] = ti_50[i] ;
    }
  }
  else if ( baud_rate == 75 )  {
    for ( int i = 0; i < 7; i++ ) {
      ti_speed[i] = ti_75[i] ;
    }
  }
  else if ( baud_rate == 100 ) {
    for ( int i = 0; i < 7; i++ ) {
      ti_speed[i] = ti_100[i] ;
    }
  }
  else if ( baud_rate == 110 ) {
    for ( int i = 0; i < 7; i++ ) {
      ti_speed[i] = ti_110[i] ;
    }
  }
  else if ( baud_rate == 255 ) {
    for ( int i = 0; i < 7; i++ ) {
      ti_speed[i] = ti_300[i] ;
    }
  }
  else
    // wenn kein gültiger EEPROM-Wert für Baudrate vorhanden ist, dann mit 45 Baud starten.
  {
    EEPROM.write(0, 45);   // und gleich in EEPROM schreiben.
    for ( int i = 0; i < 7; i++ ) {
      ti_speed[i] = ti_45[i] ;  // mit 45 Baud starten
    }
    baud_rate = EEPROM.read(0);
  }
}

//////////////////////////////////////////////////////////////
void setup()
{
  Serial.begin(9600);      // open the serial port at 9600 bps:
  Serial.println("RTTY-Start");
  lcd.begin(16, 2);                                                       // oder 16x2 ?  
  pinMode(10,OUTPUT);       // Backlight "ON"
  digitalWrite(10, 1);   
  lcd.clear();

  pinMode(11, INPUT_PULLUP); // Taster für Baudrate-Auswahl              // #11  Baudrate 
  pinMode(12, INPUT_PULLUP); // Taster für Normal/Inverse                // #12  Normal/Invers 
  pinMode(15, INPUT_PULLUP); // Eingangs-Signal                          // #A1  Input von LM567

  init_baud_timer();

  inverse = EEPROM.read(1); // Signallage ist in EEPROM Adr. 1
  if ( inverse != 0 && inverse != 1 ) {
    inverse = 0;
    EEPROM.write(1, inverse);
  }
  fig = 0;                    // mit Buchstaben-Ebene beginnen

  lcd.setCursor(0, 0);
  lcd.print("RTTY-Dspl. DD0CW" );
  lcd.setCursor(0, 1);
  lcd.print("Baud:");
  Serial.print("Baud:");
  if ( baud_rate != 255 )
  {
    lcd.print(baud_rate);
    Serial.print(baud_rate);
  }
  else {
    lcd.print("300");
    Serial.print("300");
  }

  lcd.setCursor(11, 1);
  if ( inverse == 0) {
    lcd.print("norm.");
    Serial.println("normal");
  }
  else {
    lcd.print("invs.");
    Serial.println("inv");
  }
  delay(3000);
  lcd.clear();

  FlexiTimer2::set(1, timer_interrupt);
  FlexiTimer2::start();
}

////////////////////////////////////////////////////////////
void loop()
{ static byte    x;
  static byte    y;
  static char    prev_ch; // vorhergehendes Zeichen
  static char    ch;
  static char    c[61];
  uint8_t        baudot;
  int            i;

  if ( sw_11 == true ) {
    switch (baud_rate) {
      case 45:  baud_rate =  50; break;
      case 50:  baud_rate =  75; break;
      case 75:  baud_rate = 100; break;
      case 100: baud_rate = 110; break;
      case 110: baud_rate = 255; break;
      case 255: baud_rate =  45; break;
      default:  baud_rate =  50;
    }

    EEPROM.write(0, baud_rate);
    init_baud_timer();

    lcd.clear();
    lcd.setCursor(0, 1);
    lcd.print("Baud:");
    if ( baud_rate != 255 )
    {
      lcd.print(baud_rate);
      Serial.print(baud_rate);
      Serial.print("Bd. ");
    }
    else {
      lcd.print("300");
      Serial.print("300");
      Serial.print("Bd. ");
     }

    delay(500);
    lcd.clear();
    fig = x = y = i = 0;
    sw_11 = false; return;
  }

  if ( sw_12 == true ) {
    lcd.clear();

    inverse = EEPROM.read(1);
    if ( inverse == 0 ) {
      inverse = 1;
    }
    else {
      inverse = 0;
    }

    EEPROM.write(1, inverse);
    lcd.setCursor(11, 1);
    if (inverse == 0) {
      lcd.print("norm.");
      Serial.println(" norm ");
    }
    else {
      lcd.print("invs.");
      Serial.println(" inv ");
    }
    delay(500);
    lcd.clear();
    fig = x = y = i = 0;
    sw_12 = false; return;
  }

  baudot = baud_buff[0];

  if (baudot != 0)
  {
    for ( i = 0; i < 4; i++ ) {
      baud_buff[i] = baud_buff[i + 1];
    }
    baud_buff[4] = 0;
    ch = '\0';
    if (baudot == B11111) {
      fig = 0;  //LETTERS
    }
    else if (baudot == B11011) {
      fig = 1;  //FIGURES
    }
    else if (baudot == B01000) {
      ch = '\r';  //CR
    }
    else if (baudot == B00010) {
      ch = '\n ';  //LF
    }
    else if (baudot == B00100) {
      ch = ' '; //SPACE
    }
    if (fig == 0)
    {
      if (baudot == B00011) {
        ch = 'A';
      }
      else if (baudot == B11001) {
        ch = 'B';
      }
      else if (baudot == B01110) {
        ch = 'C';
      }
      else if (baudot == B01001) {
        ch = 'D';
      }
      else if (baudot == B00001) {
        ch = 'E';
      }
      else if (baudot == B01101) {
        ch = 'F';
      }
      else if (baudot == B11010) {
        ch = 'G';
      }
      else if (baudot == B10100) {
        ch = 'H';
      }
      else if (baudot == B00110) {
        ch = 'I';
      }
      else if (baudot == B01011) {
        ch = 'J';
      }
      else if (baudot == B01111) {
        ch = 'K';
      }
      else if (baudot == B10010) {
        ch = 'L';
      }
      else if (baudot == B11100) {
        ch = 'M';
      }
      else if (baudot == B01100) {
        ch = 'N';
      }
      else if (baudot == B11000) {
        ch = 'O';
      }
      else if (baudot == B10110) {
        ch = 'P';
      }
      else if (baudot == B10111) {
        ch = 'Q';
      }
      else if (baudot == B01010) {
        ch = 'R';
      }
      else if (baudot == B00101) {
        ch = 'S';
      }
      else if (baudot == B10000) {
        ch = 'T';
      }
      else if (baudot == B00111) {
        ch = 'U';
      }
      else if (baudot == B11110) {
        ch = 'V';
      }
      else if (baudot == B10011) {
        ch = 'W';
      }
      else if (baudot == B11101) {
        ch = 'X';
      }
      else if (baudot == B10101) {
        ch = 'Y';
      }
      else if (baudot == B10001) {
        ch = 'Z';
      }
    }
    if (fig == 1)
    {
      if (baudot == B00011) {
        ch = '-';
      }
      else if (baudot == B11001) {
        ch = '?';
      }
      else if (baudot == B01110) {
        ch = ':';
      }
      else if (baudot == B01001) {  } // Wer da
      else if (baudot == B00001) {
        ch = '3';
      }
      else if (baudot == B01101) {         }
      else if (baudot == B11010) {         }
      else if (baudot == B10100) {         }
      else if (baudot == B00110) {
        ch = '8';
      }
      else if (baudot == B01011) {
        ch = '\a';  // Klingel
      }
      else if (baudot == B01111) {
        ch = '(';
      }
      else if (baudot == B10010) {
        ch = ')';
      }
      else if (baudot == B11100) {
        ch = '.';
      }
      else if (baudot == B01100) {
        ch = ',';
      }
      else if (baudot == B11000) {
        ch = '9';
      }
      else if (baudot == B10110) {
        ch = '0';
      }
      else if (baudot == B10111) {
        ch = '1';
      }
      else if (baudot == B01010) {
        ch = '4';
      }
      else if (baudot == B00101) {
        ch = '\'';
      }
      else if (baudot == B10000) {
        ch = '5';
      }
      else if (baudot == B00111) {
        ch = '7';
      }
      else if (baudot == B11110) {
        ch = '=';
      }
      else if (baudot == B10011) {
        ch = '2';
      }
      else if (baudot == B11101) {
        ch = '/';
      }
      else if (baudot == B10101) {
        ch = '6';
      }
      else if (baudot == B10001) {
        ch = '+';
      }
    }
    if (ch != '\0')

    { if ( ( ch == '\r' || ch == '\n' ) and ( prev_ch != '\r' && prev_ch != '\n' ) )
        // wenn CR oder LF, dann neue Zeile auf seriellem Terminal, aber nur 1 mal
      {
        Serial.println();
      }
      // Keine CR oder LF ausgeben auf seriell, und auch keine Spaces am Zeilen-Anfang
      else if ( ch != '\r' && ch != '\n' )
      { if ( ch == ' ' && ( prev_ch == '\r' || prev_ch == '\n' ) ) {
          prev_ch = ch; return;
        }
        else {
          Serial.print(ch); // den Rest auf seriellem Terminal ausgeben
        }
      }

      if ( ( ch == ' ' || ch == '\r' || ch == '\n' ) and
           ( prev_ch == ' ' || prev_ch == '\r' || prev_ch == '\n' )) { // nicht mehrere Spaces hintereinander auf LCD anzeigen
        prev_ch = ch; return;
      }

      prev_ch = ch;

      if ( ch == '\r' || ch == '\n' ) { // auf LCD wird CR oder LF bzw. mehrere zu einzelnem Space
        ch = ' ';
      }

      lcd.print(ch);
      c[x] = ch;
      x++;
      if (x == 16 && y == 0)
      {
        x = 0;
        y = 1;
      }
      if (x == 16 && y == 1)
      {
        lcd.clear();
        for (i = 0; i < 16; i++)
        {
          lcd.print(c[i]);
        }
        x = 0;
        y = 1;
      }
      lcd.setCursor(x, y);
    }
  }
}

